use crate::error::{BrokenPipeError, ConnectError, Error, MapBrokenPipe};
use crate::hostname::Hostname;
use crate::packet::try_read_packet;
use crate::packet::{Packet, PacketData};
use crate::state::ConnectionState;
use crate::stream::Stream;

use kiss_tnc::Tnc;
use std::time::{Duration, Instant};
use tokio::io::{AsyncRead, AsyncWrite};
use tokio::net::ToSocketAddrs;
use tokio::sync::{mpsc, mpsc::error::TryRecvError};

pub struct Client {
	us: Hostname,
	timeout: Duration,
	max_packet_size: usize,
	max_retries: usize,
	baud_rate: usize,

	// sends packets to the Tnc (mpsc)
	packet_sender: mpsc::UnboundedSender<Vec<u8>>,
	// sends channels which packets should be sent through
	channel_sender: mpsc::Sender<mpsc::Sender<Packet>>,
}

impl Client {
	/// Create a new client
	/// Used for initiating connections and sending pings
	pub async fn new_tcp<A: ToSocketAddrs>(
		addr: A,
		us: Hostname,
		timeout: Duration,
		max_packet_size: usize,
		max_retries: usize,
		baud_rate: usize,
	) -> Result<Client, ConnectError> {
		let tnc = Tnc::connect_tcp(addr).await?;

		Client::new(tnc, us, timeout, max_packet_size, max_retries, baud_rate).await
	}

	pub async fn new<
		R: AsyncRead + Unpin + Send + 'static,
		W: AsyncWrite + Unpin + Send + 'static,
	>(
		tnc: Tnc<R, W>,
		us: Hostname,
		timeout: Duration,
		max_packet_size: usize,
		max_retries: usize,
		baud_rate: usize,
	) -> Result<Client, ConnectError> {
		let (packet_sender, mut packet_receiver) = mpsc::unbounded_channel::<Vec<_>>();
		let (channel_sender, mut channel_receiver) = mpsc::channel(1);

		let (mut tnc_read, mut tnc_write) = tnc.split();

		// take incoming frames and stuff them in a channel
		{
			let packet_sender = packet_sender.clone();

			tokio::task::spawn(async move {
				let mut channels: Vec<mpsc::Sender<Packet>> = vec![];
				while let Ok((_, data)) = tnc_read.read_frame().await {
					loop {
						match channel_receiver.try_recv() {
							Ok(ch) => channels.push(ch),
							Err(TryRecvError::Disconnected) => {
								// end thread
								return;
							}
							Err(TryRecvError::Empty) => {
								// break loop
								break;
							}
						}
					}

					if let Some(packet) = Packet::from_bytes(&data) {
						if packet.to == us {
							for i in (0..channels.len()).rev() {
								if channels[i].send(packet.clone()).await.is_err() {
									channels.remove(i);
								}
							}

							// handle pings
							if packet.data == PacketData::Ping {
								// send pong
								let mut buf = vec![];
								Packet::new(
									us,
									packet.from,
									(packet.seq_num % 255) + 1,
									PacketData::Pong,
								)
								.to_bytes(&mut buf);
								if packet_sender.send(buf).is_err() {
									break;
								}
							}
						}
					}
				}
			});
		}

		// process outgoing frames
		{
			tokio::task::spawn(async move {
				while let Some(data) = packet_receiver.recv().await {
					if tnc_write.send_frame(&data).await.is_err() {
						break;
					}
					let _ = tnc_write.flush().await;
				}
			});
		}

		Ok(Self {
			us,
			timeout,
			max_packet_size,
			max_retries,
			baud_rate,
			packet_sender,
			channel_sender,
		})
	}

	/// Connects to the specified server.
	/// Returns a stream which can be used to send/receive data.
	pub async fn connect(&self, to: Hostname) -> Result<Stream, ConnectError> {
		let (s1, s2) = Stream::new_entangled();
		let (packet_sender, packet_receiver) = mpsc::channel(1);
		self.channel_sender
			.send(packet_sender)
			.await
			.map_broken_pipe()?;

		let mut client = ClientReader {
			state: ConnectionState::new(
				self.timeout,
				self.max_packet_size - crate::packet::PACKET_OVERHEAD_BYTES,
				self.max_retries,
				self.baud_rate,
				s1,
				self.packet_sender.clone(),
				self.us,
				to,
			),
			packet_receiver,
		};

		client.actually_connect().await?;

		// process incoming frames
		{
			tokio::task::spawn(async move {
				let _ = client.process_forever().await;
			});
		}

		Ok(s2)
	}

	/// Pings a host
	pub async fn ping(&self, to: Hostname) -> Result<bool, BrokenPipeError> {
		let id = rand::random();
		let next = (id % 255) + 1;

		let (packet_sender, mut packet_receiver) = mpsc::channel(1);
		self.channel_sender
			.send(packet_sender)
			.await
			.map_broken_pipe()?;

		let mut buf = vec![];
		Packet::new(self.us, to, id, PacketData::Ping).to_bytes(&mut buf);
		self.packet_sender.send(buf).map_broken_pipe()?;

		let max_time = Instant::now() + self.timeout;
		while Instant::now() < max_time {
			if let Some(packet) = try_read_packet(&mut packet_receiver)? {
				if packet == Packet::new(to, self.us, next, PacketData::Pong) {
					// Ping success
					return Ok(true);
				}
			}

			tokio::time::sleep(Duration::from_millis(10)).await;
		}

		// No response
		Ok(false)
	}
}

pub struct ClientReader {
	state: ConnectionState,
	packet_receiver: mpsc::Receiver<Packet>,
}

impl ClientReader {
	async fn process_forever(&mut self) -> Result<(), Error> {
		loop {
			let r =
				tokio::time::timeout(Duration::from_millis(50), self.packet_receiver.recv()).await;

			match r {
				Ok(Some(packet)) => {
					if packet.to == self.state.us && packet.from == self.state.them {
						self.process_packet(packet).await?;
					}
				}

				Ok(None) => {
					// channel died
					// kill our thread
					return Err(Error::BrokenPipe);
				}

				Err(_) => {
					// no packet within timeout
					// loop as usual
				}
			}

			self.state.process().await?;
		}
	}

	async fn process_packet(&mut self, packet: Packet) -> Result<(), Error> {
		self.state.incoming_packet(packet).await
	}

	async fn actually_connect(&mut self) -> Result<(), ConnectError> {
		let mut connected = false;
		for _ in 0..self.state.max_retries {
			match self.try_establish_connection().await {
				Ok(()) => {
					connected = true;
					break;
				}

				Err(ConnectError::ConnectionFailure) => {} // try again
				Err(e) => {
					return Err(e);
				}
			}
		}

		if !connected {
			return Err(ConnectError::ConnectionFailure);
		}

		Ok(())
	}

	async fn try_establish_connection(&mut self) -> Result<(), ConnectError> {
		// Send connect packet
		let mut buf = vec![];
		Packet::new(
			self.state.us,
			self.state.them,
			self.state.seq_inc(),
			PacketData::Connect,
		)
		.to_bytes(&mut buf);

		self.state.packet_sender.send(buf).map_broken_pipe()?;

		let max_time = Instant::now() + self.state.timeout;
		while Instant::now() < max_time {
			if let Some(packet) = try_read_packet(&mut self.packet_receiver)? {
				if packet.from == self.state.them
					&& packet.to == self.state.us
					&& packet.data == PacketData::ConnectAck
				{
					// Connection established
					return Ok(());
				}
			}

			tokio::time::sleep(Duration::from_millis(10)).await;
		}

		Err(ConnectError::ConnectionFailure)
	}
}
