use crate::error::{BrokenPipeError, Error, MapBrokenPipe};
use crate::hostname::Hostname;
use crate::packet::try_read_packet;
use crate::packet::{Packet, PacketData};
use crate::state::ConnectionState;
use crate::stream::Stream;

use kiss_tnc::tnc::Tnc;
use std::collections::HashMap;
use std::time::{Duration, Instant};
use tokio::io::{AsyncRead, AsyncWrite};
use tokio::net::ToSocketAddrs;
use tokio::sync::{
	mpsc,
	mpsc::{Receiver, Sender, UnboundedSender},
};

pub struct Server {
	stream_receiver: Receiver<(Hostname, Stream)>,

	// sends channels which packets should be sent through
	// only used for pings right now
	channel_sender: Sender<Sender<Packet>>,

	// packets to be sent over the air
	// also only for pings right now
	packet_sender: UnboundedSender<Vec<u8>>,

	hostname: Hostname,
	timeout: Duration,
}

impl Server {
	// TODO builder

	pub async fn new_tcp<A: ToSocketAddrs>(
		addr: A,
		hostname: Hostname,
		timeout: Duration,
		max_packet_size: usize,
		max_retries: usize,
		baud_rate: usize,
	) -> Result<Self, std::io::Error> {
		let tnc = Tnc::connect_tcp(addr).await?;

		Self::new(
			tnc,
			hostname,
			timeout,
			max_packet_size,
			max_retries,
			baud_rate,
		)
		.await
	}

	pub async fn new<
		R: AsyncRead + Unpin + Send + 'static,
		W: AsyncWrite + Unpin + Send + 'static,
	>(
		tnc: Tnc<R, W>,
		hostname: Hostname,
		timeout: Duration,
		max_packet_size: usize,
		max_retries: usize,
		baud_rate: usize,
	) -> Result<Self, std::io::Error> {
		let (stream_sender, stream_receiver) = mpsc::channel(1);
		let (packet_sender, mut packet_receiver) = mpsc::unbounded_channel();
		let (channel_sender, mut channel_receiver) = mpsc::channel::<mpsc::Sender<Packet>>(1);
		let (received_packet_tx, mut received_packet_rx) = mpsc::channel(1);

		let (mut tnc_read, mut tnc_write) = tnc.split();

		// take incoming frames and stuff them in a channel
		{
			tokio::task::spawn(async move {
				let mut channel = None;
				while let Ok((_, data)) = tnc_read.read_frame().await {
					if let Ok(ch) = channel_receiver.try_recv() {
						channel = Some(ch);
					}

					if let Some(packet) = Packet::from_bytes(&data) {
						if let Some(ch) = &channel {
							if ch.send(packet.clone()).await.is_err() {
								channel = None;
							}
						}

						if received_packet_tx.send(packet).await.is_err() {
							// Channel closed. End this thread
							break;
						};
					}
				}
			});
		}

		// processes incoming frames
		{
			let packet_sender = packet_sender.clone();
			tokio::task::spawn(async move {
				let mut reader = ServerReader::new(
					hostname,
					stream_sender,
					packet_sender,
					timeout,
					max_packet_size,
					max_retries,
					baud_rate,
				);

				loop {
					let r =
						tokio::time::timeout(Duration::from_millis(50), received_packet_rx.recv())
							.await;

					match r {
						Ok(Some(packet)) => {
							if packet.to == hostname {
								if let Err(BrokenPipeError) = reader.process_packet(packet).await {
									// channel died
									// kill our thread
									break;
								}
							}
						}

						Ok(None) => {
							// channel died
							// kill our thread
							break;
						}

						Err(_) => {
							// no packet within timeout
							// loop as usual
						}
					}

					let mut disconnected_hosts = vec![];
					for (host, state) in &mut reader.connections {
						match state.process().await {
							Ok(()) => {}

							Err(Error::Disconnected) => {
								if let Err(BrokenPipeError) =
									state.disconnect(hostname, *host).await
								{
									// channel died
									// kill our thread
									break;
								}

								disconnected_hosts.push(*host);
							}

							Err(Error::BrokenPipe) => {
								// channel died
								// kill our thread
								break;
							}
						}
					}

					for host in disconnected_hosts {
						reader.connections.remove(&host);
					}
				}
			});
		}

		// processes outgoing frames
		tokio::task::spawn(async move {
			loop {
				while let Some(data) = packet_receiver.recv().await {
					if tnc_write.send_frame(&data).await.is_err() {
						break;
					}
					let _ = tnc_write.flush().await;
				}
			}
		});

		Ok(Self {
			stream_receiver,
			channel_sender,
			packet_sender,
			hostname,
			timeout,
		})
	}

	/// Accepts a connection from a client.
	/// Returns an error if the channel is disconnected (i.e. something went wrong with the TNC connection).
	///
	/// NOTE: It is critical that this function is called regularly, even if the result is thrown away!
	/// Otherwise, the streams will pile up and incoming packets will no longer be accepted properly (even pings!)
	pub async fn accept(&mut self) -> Result<(Hostname, Stream), BrokenPipeError> {
		self.stream_receiver.recv().await.map_broken_pipe()
	}

	/// Pings a host
	pub async fn ping(&self, to: Hostname) -> Result<bool, BrokenPipeError> {
		let id = rand::random();
		let next = (id % 255) + 1;

		let (packet_sender, mut packet_receiver) = mpsc::channel(1);
		self.channel_sender
			.send(packet_sender)
			.await
			.map_broken_pipe()?;

		let mut buf = vec![];
		Packet::new(self.hostname, to, id, PacketData::Ping).to_bytes(&mut buf);
		self.packet_sender.send(buf).map_broken_pipe()?;

		let max_time = Instant::now() + self.timeout;
		while Instant::now() < max_time {
			if let Some(packet) = try_read_packet(&mut packet_receiver)? {
				if packet == Packet::new(to, self.hostname, next, PacketData::Pong) {
					// Ping success
					return Ok(true);
				}
			}

			tokio::time::sleep(Duration::from_millis(10)).await;
		}

		// No response
		Ok(false)
	}
}

struct ServerReader {
	hostname: Hostname,
	stream_sender: Sender<(Hostname, Stream)>,
	packet_sender: UnboundedSender<Vec<u8>>,
	connections: HashMap<Hostname, ConnectionState>,

	// Needed to pass the settings to the ConnectionState
	timeout: Duration,
	max_packet_size: usize,
	max_retries: usize,
	baud_rate: usize,
}

impl ServerReader {
	fn new(
		hostname: Hostname,
		stream_sender: Sender<(Hostname, Stream)>,
		packet_sender: UnboundedSender<Vec<u8>>,
		timeout: Duration,
		max_packet_size: usize,
		max_retries: usize,
		baud_rate: usize,
	) -> Self {
		Self {
			hostname,
			stream_sender,
			packet_sender,
			connections: HashMap::new(),

			timeout,
			max_packet_size,
			max_retries,
			baud_rate,
		}
	}

	async fn process_packet(&mut self, packet: Packet) -> Result<(), BrokenPipeError> {
		match &packet.data {
			PacketData::Connect => self.connection_from(packet.from).await?,
			PacketData::Ping => self.pong(packet.from, packet.seq_num).await?,
			_ => {}
		}

		let from = packet.from;
		if let Some(conn) = self.connections.get_mut(&from) {
			match conn.incoming_packet(packet).await {
				Ok(()) => {}
				Err(Error::Disconnected) => {
					self.connections.remove(&from);
				}
				Err(Error::BrokenPipe) => return Err(BrokenPipeError),
			}
		}

		Ok(())
	}

	async fn pong(&mut self, to: Hostname, id: u8) -> Result<(), BrokenPipeError> {
		let mut buf = vec![];
		Packet::new(self.hostname, to, (id % 255) + 1, PacketData::Pong).to_bytes(&mut buf);
		self.packet_sender.send(buf).map_broken_pipe()?;

		Ok(())
	}

	async fn connection_from(&mut self, from: Hostname) -> Result<(), BrokenPipeError> {
		let (s1, s2) = Stream::new_entangled();

		let mut state = ConnectionState::new(
			self.timeout,
			self.max_packet_size - crate::packet::PACKET_OVERHEAD_BYTES,
			self.max_retries,
			self.baud_rate,
			s1,
			self.packet_sender.clone(),
			self.hostname,
			from,
		);

		// Send ACK
		let mut buf = vec![];
		Packet::new(self.hostname, from, state.seq_inc(), PacketData::ConnectAck)
			.to_bytes(&mut buf);
		self.packet_sender.send(buf).map_broken_pipe()?;

		// If we're already connected, reset the connection

		self.connections.insert(from, state);
		self.stream_sender
			.send((from, s2))
			.await
			.map_broken_pipe()?;

		Ok(())
	}
}
